#ifndef PDG_CLOOG_H
#define PDG_CLOOG_H

#include <cloog/cloog.h>
#undef value_compare
#include "pdg.h"

char **cloog_util_generate_names(int nb_items, char * prefix);
char **cloog_util_parameter_names(const seq<pdg::parameter>& params);
clast_stmt *cloog_util_create_clast(CloogProgram *p, CloogOptions *options);

/* Replace compound conditionals by cascading conditionals.
 * I.e., 
 *	    if (a && b)
 * =>
 *	    if (a)
 *		if(b)
 *
 * Also replace compound upper bounds by a simple upper bound
 * and a set of conditionals.
 */
void cloog_util_simple_guards(clast_stmt *s);

/* Hoist ifs that are common to several adjacent statements.
 *
 * The code assumes that cloog_util_simple_guards has been called first.
 */
clast_stmt *cloog_util_hoist_ifs(clast_stmt *s);

struct cloog_program_builder {
    CloogProgram *p;
    CloogDomainList *scatterings;

    cloog_program_builder();
    cloog_program_builder(Polyhedron *ctx);
    cloog_program_builder(pdg::PDG *pdg);
    ~cloog_program_builder();

    void add_domain(Polyhedron *P, int nr, void *data = NULL);
    void set_names(const seq<pdg::parameter>& params, int nscat);
    void add_scattering(Polyhedron *S);
private:
    void init();
    CloogLoop **next;
    CloogBlockList **nextBL;
    CloogDomainList **nextScat;
    int nparam;
    int niter;
};

#endif
