#include <iostream>
#include <getopt.h>
#include "pdg.h"
extern "C" {
#include <barvinok/evalue.h>
}
#include <barvinok/barvinok.h>
#include <omega.h>

extern double count_solutions(Relation& r);

using std::cerr;

typedef std::vector<Variable_ID> varvector;

Relation constraints2relation(
    int in, int out, std::vector< std::vector<int> > & c)
{
    Relation r (in, out);
    assert(c.size() >= 1);
    int	dim = c[0].size();
    varvector vars;

    F_Exists *e = r.add_exists();
    F_And *base = e->add_and();

    for (int j = 1; j <= r.n_inp(); ++j)
	vars.push_back(r.input_var(j));
    for (int j = 1; j <= r.n_out(); ++j)
	vars.push_back(r.output_var(j));
    for (int i = 0; i < dim-2 - in - out; ++i)
	vars.push_back(e->declare());

    for (int i = 0; i < c.size(); ++i) {
	Constraint_Handle h;
	if (c[i][0]) 
	    h = base->add_GEQ();
	else
	    h = base->add_EQ();
	for (int j = 1; j < dim-1; ++j)
	    h.update_coef(vars[j-1], c[i][j]);
	h.update_const(c[i][dim-1]);
    }
    r.finalize();
    return r;
}

#ifdef GNUMP
#define value_to_int(v) ((int)mpz_get_si(v))
#else
#define value_to_int(v) ((int)v)
#endif

using pdg::PDG;

struct option options[] = {
    { "input",	required_argument,  0,  'i' },
    { "output",	required_argument,  0,  'o' },
    { 0, 0, 0, 0 }
};

int main(int argc, char **argv)
{
    char *input = NULL, *output = NULL;
    FILE *in, *out;
    int c, ind = 0;
    PDG *main_pdg;

    while ((c = getopt_long(argc, argv, "io", options, &ind)) != -1) {
	switch (c) {
	case 'i':
	    input = optarg;
	    break;
	case 'o':
	    output = optarg;
	    break;
	}
    }
    in = input ? fopen(input, "r") : stdin;
    assert(in);
    main_pdg = yaml::Load<PDG>(in);
    if (input)
	fclose(in);

    for (int i = 0; i < main_pdg->nodes.size(); ++i) {
	pdg::statement *s = main_pdg->nodes[i]->statement;
	for (int j = 0; j < s->accesses.size(); ++j) {
	    for (int k = 0; k < s->accesses[j]->backward.size(); ++k) {
		pdg::reuse_pair *rp = s->accesses[j]->backward[k];
		int n = rp->map->constraints.size();
		int counts[n];
		int count = 0;
		Polyhedron *D(*rp->map);
		if (!D) {
		    rp->count = 0;
		    continue;
		}

		int exist = D->Dimension - rp->map->input - rp->map->output;
		//fprintf(stderr, "%d %d %d %d\n", i, j, k, exist);
		//Polyhedron_Print(stderr, P_VALUE_FMT, D);
		int l = n-1;
		for (Polyhedron *P = D; P; P = P->next, --l) {
		    Polyhedron *N = P->next;
		    P->next = NULL;
		    evalue *ev = barvinok_enumerate_e(P, exist, 0, 0);
		    //fprintf(stderr, "%d\n", value_to_int(ev->x.p->arr[1].x.n));
		    count += counts[l] = value_to_int(ev->x.p->arr[1].x.n);
		    P->next = N;
		}
		for (int l = rp->map->constraints.size()-1; l >=0; --l) {
		    Relation s = constraints2relation
			(rp->map->input, rp->map->output,
				rp->map->constraints[l]->el);
		    /*
		    int c = (int)(count_solutions(s)+0.5);
		    if(counts[l] != c) {
			s.print(stderr);
			cerr << s << std::endl;
			cerr << "Parker: " << c << std::endl;
			cerr << "Barvinok: " << counts[l] << std::endl;
		    }
		    */
		}
		rp->count = count;
	    }
	}
    }

    main_pdg->Dump();
    main_pdg->free();
    delete main_pdg;
}
