#include <iostream>
#include "pdgutil.h"

#include "yaml.h"
#include "pdg.h"

using namespace std;
using namespace pdg;

int main(int argc, char * argv[])
{
    PDG *pdg;
    pdg = yaml::Load<PDG>(stdin);

    std::vector<Global_Var_Decl *> globals = pdg::extract_globals(pdg);

    for (int i = 0; i < pdg->dependences.size(); ++i) {
	pdg::dependence *dep = pdg->dependences[i];
	if (dep->type == pdg::dependence::uninitialized)
	    continue;
	pdg::node* from = dep->from;
	pdg::node* to = dep->to;
	int from_access = dep->from_access ? dep->from_access->nr : -1;
	int to_access = dep->to_access ? dep->to_access->nr : -1;
	const char *from_f = from->statement->top_function ?
			     from->statement->top_function->name->s.c_str() :
			     "(none)";
	const char *to_f = to->statement->top_function ?
			   to->statement->top_function->name->s.c_str() : "(none)";
	pdg::UnionMap *rel = dep->relation;
	Relation R = pdg::map2relation(rel, globals);
	printf("%d,%d (%s,line %d) --(%s)--> %d,%d (%s,line %d): ",
		from->nr, from_access, from_f, from->statement->line,
		dep->array->name->s.c_str(),
		to->nr, to_access, to_f, to->statement->line
		);
	R.simplify();
	R.print();
    }

    pdg->free();
    delete pdg;

    return 0;
}
